package org.apache.nutch.parse.feed;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.util.StringUtils;
import org.apache.nutch.metadata.Metadata;
import org.apache.nutch.net.protocols.Response;
import org.apache.nutch.parse.Outlink;
import org.apache.nutch.parse.Parse;
import org.apache.nutch.parse.ParseData;
import org.apache.nutch.parse.ParseImpl;
import org.apache.nutch.parse.ParseStatus;
import org.apache.nutch.parse.Parser;
import org.apache.nutch.parse.ParserFactory;
import org.apache.nutch.parse.ParserNotFound;
import org.apache.nutch.protocol.Content;
import org.apache.nutch.util.NutchConfiguration;
import org.xml.sax.InputSource;

import com.sun.syndication.feed.synd.SyndCategory;
import com.sun.syndication.feed.synd.SyndContent;
import com.sun.syndication.feed.synd.SyndEntry;
import com.sun.syndication.feed.synd.SyndFeed;
import com.sun.syndication.feed.synd.SyndPerson;
import com.sun.syndication.io.SyndFeedInput;

/**
 * @author Doğacan Güney
 */
public class FeedParser implements Parser {
  
  public static final Log LOG = LogFactory.getLog("org.apache.nutch.parse.feed");
  
  private Configuration conf;
  
  private ParserFactory parserFactory;
  
  public Map<String, Parse> getParse(Content content) {
    InputSource input = new InputSource(
        new ByteArrayInputStream(content.getContent()));

    SyndFeed feed = null;
    Map<String, Parse> parseMap = new HashMap<String, Parse>();
    try {
      SyndFeedInput feedInput = new SyndFeedInput();
      feed = feedInput.build(input);
    } catch (Exception e) {
      // return empty parse
      if (LOG.isWarnEnabled()) {
        LOG.warn("Parse failed: " + StringUtils.stringifyException(e));
      }
      return new ParseStatus(e).getEmptyParseMap(getConf());
    }
    
    List entries = feed.getEntries();
    for (Iterator i = entries.iterator(); i.hasNext(); ) {
      SyndEntry entry = (SyndEntry) i.next();
      addToMap(parseMap, entry, content);
    }
    
    return parseMap;
  }
  
  public void addToMap(Map<String, Parse> parseMap, 
      SyndEntry entry, Content content) {
    String link = entry.getLink(), title = entry.getTitle(), text;
    Metadata parseMeta = new Metadata(), contentMeta = content.getMetadata();
    Parse parse = null;
    SyndContent description = entry.getDescription();
    
    parseMeta.set("feed", content.getUrl());
    addFields(parseMeta, contentMeta, entry);

    String contentType = contentMeta.get(Response.CONTENT_TYPE);

    if (description != null) {
      text = description.getValue();
    } else {
      List contents = entry.getContents();
      StringBuffer buf = new StringBuffer();
      for (Iterator i = contents.iterator(); i.hasNext();) {
        SyndContent syndContent = (SyndContent) i.next();
        buf.append(syndContent.getValue());
      }
      text = buf.toString();
    }

    try {
      Parser parser = parserFactory.getParsers(contentType, link)[0];
      parse = ParseImpl.getFirstParseEntry(parser.getParse(
          new Content(link, link, 
              text.getBytes(), contentType, 
              contentMeta, conf)));
      parse = new ParseImpl(parse.getText(), 
          new ParseData(parse.getData().getStatus(), title, 
              parse.getData().getOutlinks(), contentMeta));
    } catch (ParserNotFound e) {
      parse = new ParseImpl(text, 
          new ParseData(new ParseStatus(ParseStatus.SUCCESS), 
                        title, new Outlink[0], contentMeta));
    }
    
    mergeMetadata(parseMeta, parse.getData().getParseMeta());
    parseMap.put(link, parse);
  }

  private void addFields(Metadata parseMeta, Metadata contentMeta, 
      SyndEntry entry) {
    List authors = entry.getAuthors(), 
         categories = entry.getCategories();
    Date published = entry.getPublishedDate(), 
         updated = entry.getUpdatedDate();
    SyndContent description = entry.getDescription();
    String contentType = "text/plain";
    
    if (description != null) {
      contentType = description.getType();
    } else {
      // TODO: What to do if contents.size() > 1?
      List contents = entry.getContents();
      if (contents.size() > 0) {
        contentType = ((SyndContent) contents.get(0)).getType();
      }
    }
    
    if (authors != null) {
      for (Object o : authors) {
        SyndPerson author = (SyndPerson) o;
        String authorName = author.getName();
        if (checkString(authorName)) {
          parseMeta.set("author", authorName);
        }
      }
    } else {
      // getAuthors may return null if the feed is non-atom
      // if so, call getAuthor to get Dublin Core module creator.
      String authorName = entry.getAuthor();
      if (checkString(authorName))
        parseMeta.set("author", authorName);
    }

    if (categories != null) {
      for (Iterator i = categories.iterator(); i.hasNext();) {
        SyndCategory category = (SyndCategory) i.next();
        parseMeta.add("category", category.getName());
      }
    }
    if (published != null) {
      parseMeta.set("published", Long.toString(published.getTime()));
    }
    if (updated != null) {
      parseMeta.set("updated", Long.toString(updated.getTime()));
    }
    if (checkString(contentType)) {
      // ROME may return content-type as html
      if (contentType.equals("html"))
        contentType = "text/html";
      contentMeta.set(Response.CONTENT_TYPE, contentType);
    } else {
      contentMeta.set(Response.CONTENT_TYPE, "text/plain");
    }
  }
  
  private void mergeMetadata(Metadata first, Metadata second) {
    for (String name : second.names()) {
      String[] values = second.getValues(name);
      for (String value : values) {
        first.add(name, value);
      }
    }
  }
  
  private boolean checkString(String s) {
    return s != null && !s.equals("");
  }

  public void setConf(Configuration conf) {
    this.conf = conf;
    this.parserFactory = new ParserFactory(conf);
  }

  public Configuration getConf() {
    return this.conf;
  }
  
  public static void main(String[] args) throws Exception {
    //LOG.setLevel(Level.FINE);
    if (args.length != 1) {
      System.err.println("Usage: FeedParser <feed_link>");
      System.exit(1);
    }
    String name = args[0];
    String url = "file:"+name;
    Configuration conf = NutchConfiguration.create();
    FeedParser parser = new FeedParser();
    parser.setConf(conf);
    File file = new File(name);
    byte[] bytes = new byte[(int)file.length()];
    DataInputStream in = new DataInputStream(new FileInputStream(file));
    in.readFully(bytes);
    Map<String, Parse> parseMap = parser.getParse(
            new Content(url, url, bytes, "application/rss+xml", 
                        new Metadata(), conf));
    for (Entry<String, Parse> entry : parseMap.entrySet()) {
      System.out.println("key: "  + entry.getKey());
      Parse parse = entry.getValue();
      System.out.println("data: " + parse.getData());
      System.out.println("text: " + parse.getText() + "\n");
    }
  }
}
